/**
 * <p>Copyright: Copyright (c) 2005, Knowledge Science Group, University of Calgary.
 * Permission to use, copy, modify, distribute and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that copyright
 * notice and this permission notice appear in supporting documentation.  The
 * Knowledge Science Group makes no representations about the suitability of
 * this software for any purpose.  It is provided "as is" without express or
 * implied warranty.</p>
 */
package kartaCASA;

import casa.exceptions.IPSocketException;
import casa.interfaces.TransientAgentInterface;
import casa.MLMessage;
import casa.URLDescriptor;
import casa.Status;
import casa.ML;
import casa.Act;
import casa.ui.AgentUI;
import casa.LAC;

import karta.Pickler.Pickler;
import karta.command.*;

import java.awt.Container;
import javax.swing.JFrame;
import javax.swing.JInternalFrame;

import java.util.Vector;

import kartaCASA.ui.ModelAgentWindow;

/**
 * KartaModelAgent is the superclass for KartaCASAModelAgent and KartaOntologyModelAgent,
 * which will provide specific filtering behavior for model-changing commands.
 * 
 * Created on May 31, 2005
 * @author <a href="mailto:sedachv@cpsc.ucalgary.ca">Vladimir Sedach</a>
 * @version 0.11
 */
public class KartaModelAgent extends KartaAgent {
	
	boolean debugp = false;
	
	/** A list of sleeping views */
	protected Vector<PersistentViewProxy> sleeping_views_list;
	/** A list of active views */
	protected Vector<PersistentViewProxy> views;
	
	/** The fancy pants graphics window */
	ModelAgentWindow modelUI;
	
	/** Used by CommandMaker */
	protected Pickler pickler;
	
	/** Used to make commands to send to the view */
	protected CommandMaker comMaker;
	
	
	public KartaModelAgent(Qualifiers quals) throws IPSocketException {
		super(quals);
		sleeping_views_list = new Vector<PersistentViewProxy>();
		views = new Vector<PersistentViewProxy>();
		pickler = new Pickler();
		comMaker = new CommandMaker(pickler);
	}
	
	public Status handleMessage (MLMessage msg) {
		in("KartaModelAgent.handleMessage");
		
		Status stat = new Status(0);
		
		if ((msg.getParameter(ML.PERFORMATIVE).equals(ML.REQUEST)) && (isAAct(msg.getAct(), new Act(ML.ACT_EXECUTE)))) {
			stat = handleModelChangeMessage(msg);
		}
		else if ((msg.getParameter(ML.PERFORMATIVE).equals(ML.SUBSCRIBE)) && (isAAct(msg.getAct(), new Act(ACT_JOIN_MODEL)))) {
			stat = handleViewSubscribeMessage(msg);
		}
		else if ((msg.getParameter(ML.PERFORMATIVE).equals(ML.CANCEL)) && (isAAct(msg.getAct(), new Act(ACT_WITHDRAW_MODEL)))) {
			stat = handleViewUnsubscribeMessage(msg);
		}
		else if ((msg.getParameter(ML.PERFORMATIVE).equals(ML.REQUEST)) && isAAct(msg.getAct(), new Act(ACT_SET_STATE))) {
			stat = handleStateDumpMessage(msg);
		}
		else {
			if (debugp)
				System.out.println("Unknown message: " + msg);
			stat = super.handleMessage(msg);
		}
		
		out("KartaModelAgent.handleMessage");
		return stat;
	}
	
	/** This method needs to be superseded by specific model subclasses */
	protected Status handleModelChangeMessage (MLMessage msg) {
		return new Status(0);
	}
	
	/** Handles whole-state-dumping done by the view agents periodically */
	protected Status handleStateDumpMessage (MLMessage msg) {
		in("KartaModelAgent.handleStateDumpMessage");
				
		Status stat = new Status(0);
		
		if (msg.getParameter(ML.LANGUAGE).equals(KARTA_LANGUAGE)) {
			try {
				URLDescriptor from = new URLDescriptor(msg.getParameter(ML.FROM));
				PersistentViewProxy proxy = views.get(views.indexOf(from));
				proxy.setLatestState(msg.getParameter(ML.CONTENT));
			}
			catch (Throwable e) {}
		}
		
		out("KartaModelAgent.handleStateDumpMessage");
		return stat;
	}

	/** Handles a subscription request from a view agent. */
	protected Status handleViewSubscribeMessage(MLMessage msg) {
		in("KartaModelAgent.handleViewSubscribeMessage");
		
		Status stat = new Status(0);
		
		String from = msg.getParameter(ML.FROM);

		try {
			URLDescriptor who = new URLDescriptor(from);
			addProxy(who);
		}
		catch (Throwable e) {
			stat = new Status(-1);
		}
		
		out("KartaModelAgent.handleViewSubscribeMessage");
		return stat;
	}
	
	/** Makes and adds a new PersistentViewProxy to the views list. Note that it does NOT set the proxy's
	 * model-view command translator! This has to be set by inheriting methods. */
	protected PersistentViewProxy addProxy(URLDescriptor url) {
		PersistentViewProxy view_proxy = new PersistentViewProxy(url, "", 0, pickler);
		views.add(view_proxy);
		return view_proxy;
	}
	
	/** Handles a subscription cancellation request from a view agent. */
	protected Status handleViewUnsubscribeMessage(MLMessage msg) {
		in("KartaModelAgent.handleViewUnsubscribeMessage");
		
		Status stat = new Status(0);
		
		out("KartaModelAgent.handleViewUnsubscribeMessage");
		return stat;
	}	

    protected AgentUI makeDefaultGUI(String[] args, String userName) {
        in("KartaModelAgent.makeDefaultGUI");
        try {
            Container frame = null;
            if (LAC.LACinfo.desktop != null)
                frame = new JInternalFrame ();
            else
                frame = new JFrame ();
            modelUI = new ModelAgentWindow (this, frame);
            if (LAC.LACinfo.desktop != null)
                LAC.LACinfo.desktop.addAgentWindow (modelUI);
            modelUI.setSize (500, 500);
            modelUI.setName(this.getName());
            modelUI.setOpaque (true);
            modelUI.pack ();
            modelUI.show ();
        }
        catch (Exception ex) {
            println ("warning", "KartaModelAgent: Unexpected exception trying to start the window", ex);
        }
        out("KartaModelAgent.makeDefaultGUI");
        return modelUI;
    }
    
    /** Sends the given Command to all of the subscribed Views. */
    protected void sendCommand(Command com) {
    	
    	if (debugp)
    		System.out.println("Now sending views command: " + pickler.pickleCommand(com));
    	
     	int i;
    	// loop over views, sending each one command
    	for (i = 0; i < views.size(); i++) {
    		sendCommand(com, views.get(i));
    	}
    	// loop over sleeping views, putting command on stack
    	for (i = 0; i < sleeping_views_list.size(); i++) {
    		sleeping_views_list.get(i).prepCommand(com);
    	}
    }
    
    /** Sends a command to the URL of a subscribed View. */
    protected void sendCommand(Command com, URLDescriptor who) {
    	try {
    		sendCommand(com, PersistentViewProxy.findProxy(views, who));
    	}
    	catch (NullPointerException e) {}
    }
    
    /** Sends command to the given PersistentViewProxy. */
    protected void sendCommand(Command com, PersistentViewProxy view) {
    		String com_pickle = view.prepCommand(com);
    		
    		MLMessage msg = getNewMessage(ML.REQUEST, ML.ACT_EXECUTE, view.getURL());
    		msg.setParameter(ML.LANGUAGE, KARTA_LANGUAGE);
    		msg.setParameter(ML.CONTENT, com_pickle);
    		sendMessage(msg);
    }
}
