/**
 * <p>Copyright: Copyright (c) 2005, Knowledge Science Group, University of Calgary.
 * Permission to use, copy, modify, distribute and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that copyright
 * notice and this permission notice appear in supporting documentation.  The
 * Knowledge Science Group makes no representations about the suitability of
 * this software for any purpose.  It is provided "as is" without express or
 * implied warranty.</p>
 */

package kartaCASA;

import java.awt.Container;

import javax.swing.JFrame;
import javax.swing.JInternalFrame;

import casa.exceptions.IPSocketException;
import casa.LAC;
import casa.URLDescriptor;
import casa.MLMessage;
import casa.Status;
import casa.ML;
import casa.Act;
import casa.ui.AgentUI;

import karta.command.SupervisorInterface;
import karta.KartaInstance;
import karta.command.CommandInterpreter;
import karta.command.CommandInterpreterInterface;
import karta.command.Command;
import karta.util.ObjectID;

import kartaCASA.ui.SupervisorAgentWindow;

/**
 * KartaSupervisorAgent is a simple reactive agent for driving an instance of the
 * Karta user interface.
 * 
 * Created on May 30, 2005
 * @author <a href="mailto:sedachv@cpsc.ucalgary.ca">Vladimir Sedach</a>
 * @version 0.14
 */
public class KartaSupervisorAgent extends KartaAgent implements SupervisorInterface {
	
	private boolean debugp = false;
	
	/** The address of the view agent to which this Supervisor is subscribed to. */
	protected URLDescriptor viewAgent;
	/** The command interpreter for running incoming commands. */
	protected CommandInterpreterInterface kartaCommandInterpreter;
	/** The actual instance of Karta controlled by this Supervisor. */
	protected KartaInstance kartaInstance;
	/** The pretty colorful clicky graphical interface for the supervisor. */
	protected SupervisorAgentWindow supervisorUI;
	
	/** Flag that controls whether MOVE_NODE commands can be executed optimistically. */
	boolean moveNodeShortcutP;
	/** Flag that controls whether RESIZE_NODE commands can be executed optimistically. */
	boolean resizeNodeShortcutP;
	/** A counter used to keep track of which commands have been executed optimistically. */
	long commandShortcutNumber;
	
	/** The sequence number of the last command received from the View agent to which this
	 * Supervisor is subscribed to. */
	long last_command_received;
	
	public KartaSupervisorAgent(Qualifiers quals, KartaInstance kinstance) throws IPSocketException {
		super(quals);
		kartaInstance = kinstance;
		initCommonSlots();
	}
	
	public KartaSupervisorAgent(Qualifiers quals) throws IPSocketException {
		super(quals);
		kartaInstance = new KartaInstance(this);
		initCommonSlots();
	}
	
	protected void initCommonSlots() {
		kartaCommandInterpreter = new CommandInterpreter(kartaInstance);
		moveNodeShortcutP = false;
		resizeNodeShortcutP = false;
		commandShortcutNumber = 0;
	}
	
	/** Pickle and send this command to the viewAgent. */
	public void considerCommand(Command command) {
		in("KartaSupervisorAgent.considerCommand");
		
		boolean shortcut = false;
		
		command.setCommandArgument("commandOriginator", "\"" + getURL().toString() + "\"");
		
		// if it's okay to shortcut some commands, tag the command with some attribute so that it will be
		// ignored upon receipt, put tag somewhere, then after sending the command execute it right away
		
		if ((moveNodeShortcutP && (command.getType() == Command.MOVE_NODE)) ||
				(resizeNodeShortcutP && (command.getType() == Command.RESIZE_NODE))) {
			command.setCommandArgument("commandShortcutNumber", Long.toString(commandShortcutNumber++));
			shortcut = true;
		}
				
		String commandPickle = kartaInstance.getPickler().pickleCommand(command);
		
		if (debugp) 
			System.out.println("Supervisor is sending command: " + commandPickle);
			
		MLMessage msg = getNewMessage(ML.PROPOSE, ACT_CONSIDER_COMMAND, viewAgent);
		msg.setParameter(ML.LANGUAGE, KARTA_LANGUAGE);
		msg.setParameter(ML.CONTENT, commandPickle);
		sendMessage(msg);
		
		if (shortcut) {
			kartaCommandInterpreter.runCommand(command);
		}
		out("KartaSupervisorAgent.considerCommand");
	}
	
	public Status handleMessage (MLMessage msg) {
		in("KartaSupervisorAgent.handleMessage");
		
		Status stat = new Status(0);
		
		if (debugp)
			System.out.println("Supervisor received message " + msg);
		
		if ((msg.getParameter(ML.PERFORMATIVE).equals(ML.REQUEST)) && (isAAct(msg.getAct(), new Act(ML.ACT_EXECUTE)))) {
			stat = handleCommandMessage(msg);
		}
		else if ((msg.getParameter(ML.PERFORMATIVE).equals(ML.REQUEST)) && (isAAct(msg.getAct(), new Act(ACT_SET_STATE)))) {
			stat = handleSetStateMessage(msg);
		}
		else if ((msg.getParameter(ML.PERFORMATIVE).equals(ML.REQUEST)) && (isAAct(msg.getAct(), new Act(ACT_GET_STATE)))) {
			stat = handleGetStateMessage(msg);
		}
		else {
			stat = super.handleMessage(msg);
		}
		
		out("KartaSupervisorAgent.handleMessage");
		return stat;
	}
	
	/** Handles a request by our View agent to execute a specified command. */
	Status handleCommandMessage (MLMessage msg) {
		in("KartaSupervisorAgent.handleCommandMessage");
		
		// If Message is not from our view agent or is not the right language, ignore it
		if (viewAgent.equals(msg.getParameter(ML.SENDER)) && msg.getParameter(ML.LANGUAGE).equals(KARTA_LANGUAGE)) {
			String com = msg.getParameter(ML.CONTENT);
			
			if (debugp)
				System.out.println("Supervisor will now attempt to execute command: " + com);
			
			Command command = kartaInstance.getPickler().parseCommand(com);
			
			last_command_received = command.getSequenceNumber();
			
			if (!((command.getCommandArgument("commandOriginator") != null) && (getURL().equals(command.getCommandArgument("commandOriginator").toString())) && (command.getCommandArgument("commandShortcutNumber") != null) && (Long.decode((String)command.getCommandArgument("commandShortcutNumber")) <= commandShortcutNumber))) {
					kartaCommandInterpreter.runCommand(command);
			}
			
//			MLMessage reply = MLMessage.constructReplyTo(msg, getUniqueRequestID(), this.getURL());
//			//reply.setParameter(ML.CONTENT, "");
//			reply.setParameter (ML.CONTENT, CASAUtil.serialize (new Status(0)));
//			reply.setParameter (ML.LANGUAGE, "casa.*." + Status.class.getName ());
//			sendMessage(reply);
			// and then the view replies and we acknowledge, etc. etc.
			// easier just to let the commitment time out for now
			// in the future will have to modify policy so that execute does not
			// generate a commitment policy on the view side.
		}
		
		out("KartaSupervisorAgent.handleCommandMessage");
		return new Status(0);
	}
	
	/** Handle a request to update the whole concept map from the View agent. */
	Status handleSetStateMessage (MLMessage msg) {
		Status stat = new Status(-1);
		
		if (msg.getParameter(ML.LANGUAGE).equals(KARTA_LANGUAGE)) {
			if (debugp)
				System.out.println("Gotten a request to set state to: " + msg.getParameter(ML.CONTENT));
			kartaInstance.getPickler().loadMap(kartaInstance.getConceptMap(), msg.getParameter(ML.CONTENT));
			stat = new Status(0);
		}
		
		return stat;
	}
	
	/** Handle a request to save the whole concept map and give it to the view agent. */
	Status handleGetStateMessage(MLMessage msg) {
		
		if (debugp)
			System.out.println("Supervisor received request to get state.");
		
		Command dump_command = kartaInstance.getCommandMaker().dumpMapState(kartaInstance.getConceptMap(), last_command_received);
		String commandPickle = kartaInstance.getPickler().pickleCommand(dump_command);
		
		MLMessage reply = MLMessage.constructReplyTo(msg, "", getURL());
		reply.setParameter(ML.LANGUAGE, KARTA_LANGUAGE);
		reply.setParameter(ML.CONTENT, commandPickle);
		sendMessage(reply);
		
		return new Status(0);
	}
	
	protected AgentUI makeDefaultGUI(String[] args, String userName) {
		in("KartaSupervisorAgent.makeDefaultGUI");
		try {
			Container frame = null;
			if (LAC.LACinfo.desktop != null)
				frame = new JInternalFrame ();
			else
				frame = new JFrame ();
			supervisorUI = new SupervisorAgentWindow (this, frame);
			if (LAC.LACinfo.desktop != null)
				LAC.LACinfo.desktop.addAgentWindow (supervisorUI);
			supervisorUI.setSize (500, 500);
			supervisorUI.setName(this.getName());
			supervisorUI.setOpaque (true);
			supervisorUI.pack ();
			supervisorUI.show ();
		} catch (Exception ex) {
			println ("warning",
					"KartaSupervisorAgent: Unexpected exception trying to start the window",
					ex);
		}
		out("KartaSupervisorAgent.makdeDefaultGUI");
		return supervisorUI;
	}
	
	public void subscribeToView(URLDescriptor view) {
		unsubscribeFromView();
		viewAgent = view;
		MLMessage msg = getNewMessage(ML.SUBSCRIBE, ACT_JOIN_VIEW, viewAgent);
		sendMessage(msg);
		// ask the view politely whether we can shortcut some commands
		// for now we just say YES!
		moveNodeShortcutP = true;
		resizeNodeShortcutP = true;
	}
	
	public void unsubscribeFromView() {
		if (viewAgent != null) {
			MLMessage msg = getNewMessage(ML.CANCEL, ACT_WITHDRAW_VIEW, viewAgent);
			sendMessage(msg);
			viewAgent = null;
		}
		kartaCommandInterpreter.runCommand(kartaInstance.getCommandMaker().clearConceptMap());
	}
	
	protected void finishRun() {
		super.finishRun();
		unsubscribeFromView();
	}
	
	public KartaInstance getKartaInstance() {
		return kartaInstance;
	}
	
	public void setKartaInstance(KartaInstance kartaInstance) {
		this.kartaInstance = kartaInstance;
	}
	
	/** Get a new Arc ID. */
	public ObjectID getArcID() {
		return ObjectID.newArc; // this tombstone value will be replaced by the networked view
	}
	
	/** Get a new Node ID. */
	public ObjectID getNodeID() {
		return ObjectID.newNode; // this tombstone value will be replaced by the networked view
	}
	
	/** Get a new StyleSheet ID. */
	public ObjectID getStyleSheetID() {
		return ObjectID.newStyleSheet;
	}
}
