/**
 * <p>Copyright: Copyright (c) 2005, Knowledge Science Group, University of Calgary.
 * Permission to use, copy, modify, distribute and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that copyright
 * notice and this permission notice appear in supporting documentation.  The
 * Knowledge Science Group makes no representations about the suitability of
 * this software for any purpose.  It is provided "as is" without express or
 * implied warranty.</p>
 */

package kartaCASA;

import casa.exceptions.IPSocketException;
import casa.URLDescriptor;
import casa.Status;
import casa.MLMessage;
import casa.Act;
import casa.ML;
import casa.LAC;
import casa.ui.AgentUI;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.TimeUnit;
import java.util.Vector;
import java.lang.InterruptedException;

import java.awt.Container;
import javax.swing.JFrame;
import javax.swing.JInternalFrame;

import karta.command.Command;
import karta.command.CommandTree;
import karta.util.ObjectID;
import karta.util.IDMapper;
import karta.Pickler.*;

import kartaCASA.interfaces.KartaCommandFilterInterface;
import kartaCASA.ui.ViewAgentWindow;

/**
 * KartaViewAgent provides view management to KartaSupervisorAgents.
 *
 * Created on May 31, 2005
 * @author <a href="mailto:sedachv@cpsc.ucalgary.ca">Vladimir Sedach</a>
 * @version 0.42
 */
public class KartaViewAgent extends KartaAgent {
	
	private Boolean debugp = false;
	
	/** The queue of commands that need to be sent to the subscribed Supervisor agents. */
	BlockingQueue<Command> commandQueue;
	/** A Daemon that periodically runs and sends commands from the commandQueue to the subscribed Supervisors. */
	Timer commandQueueDaemon;
	
	/** A Vector of subscribed Supervisor agents. */
	protected Vector<Subscriber> subscribers;
	
	/** This view's Pickler for serializing commands and assigning Object IDs. */
	protected final Pickler pickler;
	
	// Model-related stuff
	/** The address of the model to which this view is subscribed. */
	protected URLDescriptor model;
	/** The filter used to filter commands incoming from the Supervisor agents. */
	protected KartaCommandFilterInterface commandFilter;
	/** Translator from mapping Model IDs to Object/View IDs. */
	protected ModelViewCommandTranslator commandTranslator;
	
	/** The state of the concept map being worked on. */
	protected MapState currentMap;
	
	/** The pretty colorful user interface. */
	protected ViewAgentWindow viewUI;
	
	
	/** MapState is a class that holds information about the concept map, including the whole state dump, 
	 * and the commands processed since the whole state dump. */
	class MapState {
		/** The number of the latest command to be processed. */
		long currentCommandSequenceNumber;
		/** The command sequence number up until which the last_complete_view_state is valid (already contains) to. */
		long command_seq_number_last_view_valid_upto;
		/** A pickle of the whole map, good until: SEE ON BOTTOM. */
		String last_complete_view_state;
		/** The commands that have been processed afterthe last_complete_view_state was begotten. */
		Vector<Command> command_store;
		/** the number of commands in the command_store before a whole-state-save request is made. */
		int StoreSizeStateDumpInterval = 1000; 
		
		public MapState() {
			command_store = new Vector<Command>();
			currentCommandSequenceNumber = 0;
			last_complete_view_state = "";
			command_seq_number_last_view_valid_upto = 0;
		}
		
		/** Sets the command sequence number, translates IDs and does general prepping and
		 * housekeeping for a command that is about to be sent to the Supervisors. 
		 * Called when a command is popped off the waiting queue and ready to be sent. */
		public void processNextCommand(Command com) {
			com.setSequenceNumber(currentCommandSequenceNumber++);
			
			// map newly created objects (these come with ID of ObjectID.newNode or newArc) to
			// newly created IDs by the view agent's pickler
			mapCommandID(com);
			
			command_store.add(com);
			
			// code to dump a state of the whole system if the command_store gets big
			// maybe this should be moved out to it's own daemon?
			if (command_store.size() > StoreSizeStateDumpInterval) {
				// send a get state command with currentCommandSequenceNumber in a message with ACT_GET_STATE
				MLMessage msg = getNewMessage(ML.REQUEST, ACT_GET_STATE, subscribers.get(0).getURL());
				sendMessage(msg);
			}
		}
		
		/** Returns a command with the specified command sequence number, null if that command
		 * is not found (possibly because it was already erased). */
		public Command getCommand(long num) {
			long start = command_store.get(0).getSequenceNumber();
			int offset = (int)(start - num);
			Command com = null;
			try {
				com = command_store.get(offset);
			}
			catch (ArrayIndexOutOfBoundsException e) {
				return null;
			}
			return com;
		}
		
		/** Handles a newly subscribed Supervisor. Tells it to set the state and plays back the necessary
		 * commands from the command store. */
		public void subscribe (Subscriber sub) {
			MLMessage msg = getNewMessage(ML.REQUEST, ACT_SET_STATE, sub.getURL());
			msg.setParameter(ML.LANGUAGE, KARTA_LANGUAGE);
			msg.setParameter(ML.CONTENT, last_complete_view_state);
			sendMessage(msg);
			
			if (command_store.size() > 0) {
				long first_command = command_store.get(0).getSequenceNumber();
				int com_start = (int)(command_seq_number_last_view_valid_upto - first_command);
				for(int i = com_start; i < command_store.size(); i++) {
					sub.sendCommand(command_store.get(i));
				}
			}
		}
		
		/** Sets the last view state to the specified pickle, and erases all commands with lower
		 * sequence numbers than the one the state is good until. */
		public void setLastViewState(String state, long goodUpto) {
			// get sequence number (the number of the command the state is valid up to)
			// find that sequence number in command_store
			// discard all older messages
			// find model change command sequence number (the one to which this state is good up to)
			last_complete_view_state = state;
			command_seq_number_last_view_valid_upto = goodUpto;
			long first_com_num = command_store.get(0).getSequenceNumber();
			int index_remove = (int)(command_seq_number_last_view_valid_upto - first_com_num) - 1;
			// command_store.removeRange(0, index_remove);
			// oops, removeRange is protected (what are they smoking at Sun?), so we have to do this instead:
			for (int i = 0; i < index_remove; i++) {
				command_store.remove(0);
			}
		}
	}
	
	/** Subscribed contains the necessary information and utility methods for working
	 * with Supervisor agents that have subscribed to this View. */
	class Subscriber {
		/** The address of this Supervisor agent. */
		URLDescriptor url;
		/** The sequence number of the last command which has been processed by this agent. */
		long last_received_command_number;
		
		public Subscriber(URLDescriptor u) {
			url = u;
			last_received_command_number = 0;
			// then send the state, all subsequent commands to that subscriber
		}
		
		public void sendCommand(Command com) {
			sendCommand(pickler.pickleCommand(com));
		}
		
		public void sendCommand(String com) {
			MLMessage msg = getNewMessage(ML.REQUEST, ML.ACT_EXECUTE, url);
			msg.setParameter(ML.LANGUAGE, KARTA_LANGUAGE);
			msg.setParameter(ML.CONTENT, com);
			sendMessage(msg);
		}
		
		public URLDescriptor getURL() {
			return url;
		}
	}
	
	/** This class is responsible for running periodically to process commands in the
	 * command queue and send them to all the subscribed Supervisor agents. */
	class ExecuteQueuedCommand extends TimerTask {
		public void run() {
			Command com = null;
			try {
				com = commandQueue.poll(2, TimeUnit.MILLISECONDS);
			}
			catch (InterruptedException e) {}
			if (com != null) {
				currentMap.processNextCommand(com);
				
				String com_pickle = pickler.pickleCommand(com);
				// pickle command, send it to KartaSupervisorAgent with a commitment that compels it to execute that command
				for (int i = 0; i < subscribers.size(); i++) {
					subscribers.get(i).sendCommand(com_pickle);
				}
			}
		}
	}
	
	/** Creates a new view agent with the specified KartaCommandFilter to sort incoming command messages. */
	public KartaViewAgent(Qualifiers quals) throws IPSocketException {
		super(quals);
		commandQueue = new LinkedBlockingQueue<Command>();
		commandQueueDaemon = new Timer(true); // create a daemon to periodically take messages from commandQueue and execute them
		commandQueueDaemon.scheduleAtFixedRate(new ExecuteQueuedCommand(), 0, 2);
		pickler = new Pickler();
		currentMap = new MapState();
		subscribers = new Vector<Subscriber>();
		model = null;
		
		commandFilter = new PassthroughFilter();
	}
	
	/** Adds a supervisor agent that subscribes to this view. */
	void addParticipant(URLDescriptor p) {
		if (findSubscriber(p) == -1)
			subscribers.add(new	Subscriber(p));
	}
	
	/** Removes the subscription of the given supervisor agent. */
	void removeParticipant(URLDescriptor p) {
		int index = findSubscriber(p);
		if (index != -1)
			subscribers.remove(index);
	}
	
	/** Attempts to find a subscriber with the specified URL and return it's position
	 * in the Vector of subscribers. Returns -1 if no such subscriber was found. */
	int findSubscriber(URLDescriptor p) {
		for (int i = 0; i < subscribers.size(); i++) {
			if (subscribers.get(i).getURL().equals(p))
				return i;
		}
		return -1;
	}

	public Status handleMessage (MLMessage msg) {
		in("KartaViewAgent.handleMessage");
		
		if (debugp)
			System.out.println("View received message " + msg);
		
		Status stat = new Status(0);
		String performative = msg.getParameter(ML.PERFORMATIVE);
		Act act = msg.getAct();
		
		if ((performative.equals(ML.PROPOSE)) && (isAAct(act, new Act(ACT_CONSIDER_COMMAND)))) {
			stat = handleCommandMessage(msg);
		}
		else if ((performative.equals(ML.REQUEST)) && (isAAct(act, new Act(ML.ACT_EXECUTE)))) {
			stat = handleModelCommandMessage(msg);
		}
		else if ((performative.equals(ML.AGREE)) && (isAAct(act, new Act(ACT_GET_STATE)))) {
			stat = handleGetStateMessage(msg);
		}
		else if ((performative.equals(ML.SUBSCRIBE)) && (isAAct(act, new Act(ACT_JOIN_VIEW)))) {
			stat = handleSubscribeMessage(msg);
		}
		else if ((performative.equals(ML.CANCEL)) && (isAAct(act, new Act(ACT_WITHDRAW_VIEW)))) {
			stat = handleUnsubscribeMessage(msg);
		}
		else {
			if (debugp)
				System.out.println("Unknown message has performative " + performative + " and act " + act);
			stat = super.handleMessage(msg);
		}
		
		out("KartaViewAgent.handleMessage");
		return stat;
	}

	/** Assigns new ObjectIDs to appropriate command fields and sends notification to the model agent. */
	protected void mapCommandID(final Command command) {
		if (debugp)
			System.out.println("Now mapping IDs for command: " + pickler.pickleCommand(command));
		
		Command.mapCommandIDs(command, 
				new IDMapper() { 
			public Object mapID(Object sID) {
				ObjectID ID;
				if (sID instanceof ObjectID)
					ID = new ObjectID((ObjectID)sID);
				else if (sID instanceof String)
					ID = new ObjectID((String)sID);
				else
					return sID;
				
				if (ID.equals(ObjectID.newNode)) {
					ObjectID newID = pickler.getNewNodeID();
					return newID;
				}
				else if (ID.equals(ObjectID.newArc)) {
					ObjectID newID = pickler.getNewArcID();
					return newID;
				}
				else if (ID.equals(ObjectID.newStyleSheet)) {
					ObjectID newID = pickler.getNewStyleSheetID();
					return newID;
				}
				return sID;
			}});
		if (debugp)
			System.out.println("Finished mapping IDs for command: " + pickler.pickleCommand(command));
	}
	
	/** Handles commands received from supervisor agents. */
	Status handleCommandMessage (MLMessage msg) {
		in("KartaViewAgent.handleCommandMessage");
		
		Status stat = new Status(0);
		
		if (debugp)
			System.out.println("View now trying to parse command: " + msg.getParameter(ML.CONTENT));
		
		// unpickle msg.getParameter(ML.CONTENT) to command
		Command command = pickler.parseCommand(msg.getParameter(ML.CONTENT));
		
		// when receiving a message of a certain type, put it through filter and
		// then either send it up to model agent and/or stick it into commandQueue
		int com_code = commandFilter.IsCommandValid(command);
		if (com_code != -1) { // if command is valid
			if (com_code == 0) // simple command, doesn't affect state
				putCommandInQueue(command); // stick it into the queue to filter down to supervisors
			if (com_code == 1) { // if the command changes the model
				notifyStateChangingCommand(command); // handle it specially with this protected method
			}
		}
		
		out("KartaViewAgent.handleCommandMessage");
		return stat;
	}
	
	/**
	 * Called on all commands that were deemed to be model-state modifying by the
	 * command filter. Defaults to sending the command to the currently subscribed-to
	 * model agent and adding it to the command queue.
	 */
	protected void notifyStateChangingCommand (Command com) {
		putCommandInQueue(com);
		if (model != null) {
			MLMessage model_message = getNewMessage(ML.REQUEST, ML.ACT_EXECUTE, model);
			model_message.setParameter(ML.LANGUAGE, KARTA_LANGUAGE);
			model_message.setParameter(ML.CONTENT, pickler.pickleCommand(com)); // translate command to model command
			sendMessage(model_message);
		}
	}
	
	/** Maps IDs in commands received from the Model agent into ones suitable for the view
	 * and Supervisors. If a new ID is requested, makes it and adds an entry to the model-view ID
	 * translation tables. */
	protected void mapModelCommandID(final Command command) {
		if (debugp)
			System.out.println("Now mapping IDs for model command: " + pickler.pickleCommand(command));
		
		Command.mapCommandIDs(command, 
				new IDMapper() { 
			public Object mapID(Object sID) {
				ObjectID ID;
				try {
					ID = new ObjectID((String)sID);
					ObjectID newID = ID;
					if (ID.equals(ObjectID.newNode)) {
						newID = pickler.getNewNodeID();
					}
					else if (ID.equals(ObjectID.newArc)) {
						newID = pickler.getNewArcID();
					}
					if ((command.getCommandArgument("modelID") != null)) {
						String modelID = ((StringTypeWrapper)command.getCommandArgument("modelID")).getString();
						commandTranslator.MapEntityToID(modelID, newID);
						commandTranslator.MapIDToEntity(newID, modelID);
					}
					return newID;
				}
				catch (Throwable e) { // couldn't create an ID
					return sID;
				}
			}});
		if (debugp)
			System.out.println("Finished mapping IDs for model command: " + pickler.pickleCommand(command));
	}
	
	/** Handles command messages received from the model. */
	Status handleModelCommandMessage(MLMessage msg) {
			
		if (debugp)
			System.out.println("Now handling previously received model command message.");
		
		URLDescriptor from = null;
		
		try {
			from = new URLDescriptor(msg.getParameter(ML.FROM));
			if (from.equals(model) && msg.getParameter(ML.LANGUAGE).equals(KARTA_LANGUAGE)) {
				Command modcom = pickler.parseCommand(msg.getParameter(ML.CONTENT));
				mapModelCommandID(modcom);
				commandTranslator.translateCommand(modcom);
				putCommandInQueue(modcom);
				MLMessage reply = MLMessage.constructReplyTo(msg, ML.AGREE, getURL());
				sendMessage(reply);
				return new Status(0);
			}
		}
		catch (Throwable e) {
			System.out.println("Error handling model command message " + msg);
		}
		return new Status(-1);
	}
	
	/** Handles the message sent by a supervisor agent in reply to a get-state request. */
	Status handleGetStateMessage(MLMessage msg) {
		in("KartaViewAgent.handleGetStateMessage");
				
		Status stat = new Status(0);
		
		Command com = pickler.parseCommand(msg.getParameter(ML.CONTENT));
		CommandTree last_state = (CommandTree)com.getCommandArgument("state");
		Long upto = Long.decode((String)com.getCommandArgument("upto"));
		
		if (debugp)
			System.out.println("We are now handling a get state message. Command valid upto is " + upto);
		
		currentMap.setLastViewState(pickler.pickleCommandTree(last_state), upto); 
		
		// tell model to save last good state to the persistent view proxy
		if (model != null) {
			MLMessage dumpState = getNewMessage(ML.REQUEST, ACT_SET_STATE, model);
			dumpState.setParameter(ML.LANGUAGE, KARTA_LANGUAGE);
			dumpState.setParameter(ML.CONTENT, msg.getParameter(ML.CONTENT));
			sendMessage(dumpState);
		}
		
		if (debugp)
			System.out.println("Last state is now " + pickler.pickleCommandTree(last_state));
		
		out("KartaViewAgent.handleGetStateMessage");
		return stat;
	}
	
	/** Handles subscribe request from Supervisor agents. */
	Status handleSubscribeMessage(MLMessage msg) {
		in("KartaViewAgent.handleSubscribeMessage");
		
		URLDescriptor from;
		try {
			from = msg.getURLParameter(ML.FROM);
		}
		catch (Throwable e) {
			return new Status(-1);
		}
		
		subscribeSupervisor(from);
		
		Status stat = new Status(0);
		
		out("KartaViewAgent.handleSubscribeMessage");
		return stat;
	}
	
	public void subscribeSupervisor(URLDescriptor supervisor) {
		Subscriber newbie = new Subscriber(supervisor);
		
		subscribers.add(newbie);
		
		currentMap.subscribe(newbie);
	}
	
	
	/** Handles cancel subscription message from Supervisor agents. */
	Status handleUnsubscribeMessage(MLMessage msg) {
		in("KartaViewAgent.handleUnsubscribeMessage");
		
		URLDescriptor from;
		try {
			from = msg.getURLParameter(ML.FROM);
		}
		catch (Throwable e) {
			return new Status(-1);
		}
		
		subscribers.remove(from);
		
		Status stat = new Status(0);
		
		out("KartaViewAgent.handleUnsubscribeMessage");
		return stat;
	}
	
	
	/** Puts the specified command in the commandQueue to be sent to Supervisor agents. */
	protected void putCommandInQueue(Command command) {
		try {
			commandQueue.put(command);
		}
		catch (InterruptedException e) { // retry if interrupted
			putCommandInQueue(command);
		}
	}
	
	/** Sends the command with the specified sequence number to the specified agent.
	 * @return Returns 1 if the command was sent successfuly, -1 if no such command number exists (anymore). */
	int resendCommand(Subscriber sub, long seq_num) {
		in("KartaViewAgent.resendCommand");
		
		Command com = currentMap.getCommand(seq_num);
		if (com == null) {
			return -1;
		}
		
		sub.sendCommand(com);
		
		out("KartaViewAgent.resendCommand");
		return 1;
	}
	
	public URLDescriptor getModel() {
		return model;
	}
	
	public void subscribeToModel(URLDescriptor model) {
		// need to unsubscribe from current model (if any) first
		this.model = model;
		if (debugp)
			System.out.println("Subscribing to model " + model);
		// need some way to handle errors, commitments, etc.
		MLMessage msg = getNewMessage(ML.SUBSCRIBE, ACT_JOIN_MODEL, model);
		sendMessage(msg);
	}
	
	public void unsubscribeFromModel() {
		if (model != null) {
			MLMessage msg = getNewMessage(ML.CANCEL, ACT_WITHDRAW_VIEW, model);
			sendMessage(msg);
			model = null;
		}
	}
	
	protected void finishRun() {
		super.finishRun();
		unsubscribeFromModel();
		// now we need to tell all the subscribed supervisor agents to unsubscribe from us
	}
	
	protected AgentUI makeDefaultGUI(String[] args, String userName) {
		in("KartaViewAgent.makeDefaultGUI");
		try {
			Container frame = null;
			if (LAC.LACinfo.desktop != null)
				frame = new JInternalFrame ();
			else
				frame = new JFrame ();
			viewUI = new ViewAgentWindow (this, frame);
			if (LAC.LACinfo.desktop != null)
				LAC.LACinfo.desktop.addAgentWindow (viewUI);
			viewUI.setSize (500, 500);
			viewUI.setName(this.getName());
			viewUI.setOpaque (true);
			viewUI.pack ();
			viewUI.show ();
		}
		catch (Exception ex) {
			println ("warning", "KartaViewAgent: Unexpected exception trying to start the window", ex);
		}
		out("KartaViewAgent.makeDefaultGUI");
		return viewUI;
	}
	
	public KartaCommandFilterInterface getCommandFilter() {
		return commandFilter;
	}
	
	public void setCommandFilter(KartaCommandFilterInterface commandFilter) {
		this.commandFilter = commandFilter;
	}
}
