/**
 * <p>Copyright: Copyright (c) 2005, Knowledge Science Group, University of Calgary.
 * Permission to use, copy, modify, distribute and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that copyright
 * notice and this permission notice appear in supporting documentation.  The
 * Knowledge Science Group makes no representations about the suitability of
 * this software for any purpose.  It is provided "as is" without express or
 * implied warranty.</p>
 */

package kartaCASA.MASExplorer;

import java.util.Vector;
import java.util.Hashtable;

import casa.exceptions.IPSocketException;
import casa.Act;
import casa.ML;
import casa.MLMessage;
import casa.PerformDescriptor;
import casa.Status;
import casa.URLDescriptor;

import kartaCASA.KartaModelAgent;
import kartaCASA.PersistentViewProxy;
import karta.command.Command;
import karta.util.ObjectID;
import karta.util.StringEscaper;
import karta.stylesheets.*;

/**
 * The MASModelAgent is a KartaModelAgent for the MASExplorer
 * 
 * Created on June 9, 2005
 * @author <a href="mailto:sedachv@cpsc.ucalgary.ca">Vladimir Sedach</a>
 * @version 0.17
 */
public class MASModelAgent extends KartaModelAgent {
	
	boolean debugp = false;
	
	/** A list of the CDs that have been joined by this agent. */
	Vector<URLDescriptor> CDsJoined;
	
	/** Mapping of agent URLs to agent names. */
	Hashtable<String, String> agentNames;
	
	/** A list of the nodes which don't have proper names assigned to them yet. */
	Vector<String> nonames;
	
	/** A list of newly joined CDs from which this agent has not yet received a list of members. */
	Vector<String> CDMembersWaitList;
	
	public MASModelAgent(Qualifiers quals) throws IPSocketException {
		super(quals);
		CDsJoined = new Vector<URLDescriptor>();
		agentNames = new Hashtable<String, String>();
		nonames = new Vector<String>();
		CDMembersWaitList = new Vector<String>();
	}
	
	public Status handleMessage(MLMessage msg) {
		
		Status stat = new Status(0);
		
		if ((msg.getParameter(ML.PERFORMATIVE).equals(ML.AGREE)) && (isAAct(msg.getAct(), new Act(ML.ACT_GET_NAME)))) {
			stat = handleGetNameReplyMessage(msg);
		}
		else {
			if (debugp)
				System.out.println("MASModel received unknown message " + msg);
			stat = super.handleMessage(msg);
		}
		return stat;
	}
	
	protected Status handleGetNameReplyMessage(MLMessage msg) {
		if (debugp) {
			System.out.println("Gotten name mapping message " + msg);
			System.out.println("Now mapping agent URL " + msg.getParameter(ML.FROM) + " to name " + msg.getParameter(ML.CONTENT));
		}
		// map from field to contents field
		try {
			agentNames.put(msg.getParameter(ML.FROM), msg.getParameter(ML.CONTENT));
			// if the node has been created but doesn't have a proper name,
			// send out a set_name command to all the views
			if (nonames.indexOf(msg.getParameter(ML.FROM)) != -1) {
				Command name = comMaker.setName("node", StringEscaper.escapeString(msg.getParameter(ML.FROM)), msg.getParameter(ML.CONTENT));
				sendCommand(name);
				nonames.remove(nonames.indexOf(msg.getParameter(ML.FROM)));
			}
		}
		catch (Throwable e) {
			System.out.println("Error mapping agent name because " + e);
		}

		return new Status(0);
	}
	
	protected void askForName(URLDescriptor agent) {
		if (debugp)
			System.out.println("Now asking for name of agent " + agent);
		
		if (agentNames.get(agent.toString()) == null) {
			MLMessage ask = getNewMessage(ML.REQUEST, ML.ACT_GET_NAME, agent);
			sendMessage(ask);
		}
	}
	
	/** Gets the name of the specified agent. Named as it is because getAgentName returns
	 * this agent's name.
	 */
	protected String getOtherAgentName(URLDescriptor agent) {
		return getOtherAgentName(agent.toString());
	}
	
	/** Gets the name of the specified agent. Named as it is because getAgentName returns
	 * this agent's name.
	 */
	protected String getOtherAgentName(String agent) {
		String result = agentNames.get(agent);
		if (result != null)
			return result;
		nonames.add(agent);
		return agent;
	}
	
	/** Handles a message from one of the view agents requesting a change in the model. */
	protected Status handleModelChangeMessage (MLMessage msg) {
		in("KartaModelAgent.handleModelChangeMessage");
		
		Status stat = new Status(0);
		
		out("KartaModelAgent.handleModelChangeMessage");
		return stat;
	}
	
	/** Overrides the doJoinCD (that actually makes this agent join the CD) so we can keep
	 * track of which CDs this agent is subscribed to. */
	protected PerformDescriptor release_join_cd (MLMessage msg) {
		URLDescriptor cdURL = null;
		
		try {
			cdURL = new URLDescriptor(msg.getParameter(ML.FROM));
		}
		catch (Throwable e) {}
		
		if (cdURL != null) {
			CDsJoined.add(cdURL);
			CDMembersWaitList.add(msg.getParameter(ML.FROM));
			askForName(cdURL);
		}
		
		return super.release_join_cd(msg);
	}
	
	protected PerformDescriptor release_get_members (MLMessage msg) {
		PerformDescriptor ret = super.release_get_members(msg);
		
		String cdString = msg.getParameter(ML.FROM);
		URLDescriptor cdURL = null;
		
		try {
			cdURL = new URLDescriptor(cdString);
		}
		catch (Throwable e) {}
		
		if (cdURL != null) {
			if (CDMembersWaitList.contains(cdString)) {
				CDMembersWaitList.remove(cdString);
				subscribeViewsToCD(cdURL, null); // let everyone know about this
			}
		}
		return ret;
	}
    
    /** This method handles notification of agent subscription to and withdrawal from a cooperation domain. */
    protected PerformDescriptor accept_membership_change (MLMessage msg) {
    	try {
    		URLDescriptor member = new URLDescriptor (msg.getParameter(ML.CONTENT));		
    		String senseBuf = msg.getParameter(ML.SENSE);
            boolean sense = senseBuf==null?true:senseBuf.equalsIgnoreCase("true");
            Command com;
            if (sense) { // agent joined CD
            	try {
            		URLDescriptor cd = new URLDescriptor(msg.getParameter (ML.SENDER));
            		if (getMembers(cd).indexOf(member) == -1) {
            			addAgent(member, cd, null);
            		}
            	}
            	catch (Throwable e) {
            		System.out.println("MASModelAgent: error getting name of cd from sender parameter.");
            	}
            } 
            else { // agent withdrew from CD
            	try {
            		URLDescriptor cd = new URLDescriptor(msg.getParameter (ML.SENDER));
            		removeAgent(member, cd);
            	}
            	catch (Throwable e) {}
            }
    	}
    	catch (Throwable e) {} // some sort of error
    	return super.accept_membership_change(msg);
    }
    
    /** Given a URLDescriptor of a view, creates a proxy for that view and adds it to the list
     * of proxies, */
    protected PersistentViewProxy addProxy(URLDescriptor url) {
    	PersistentViewProxy proxy = super.addProxy(url);
    	proxy.setTranslator(new MASCommandTranslator());
    	return proxy;
    }
    
    /** Handles a request from a view agent to subscribe to this model. */
    protected Status handleViewSubscribeMessage(MLMessage msg) {
    	Status msg_status = super.handleViewSubscribeMessage(msg);
    	// now the view is subscribed

    	if (debugp)
    		System.out.println("A view agent has subscribed!");
    	
    	URLDescriptor view = null;
    	try {
    		view = new URLDescriptor(msg.getParameter(ML.FROM));
    	}
    	catch (Throwable e) {}
    	
    	if (view != null) {
    		// note that the code below will totally crap out when you have one agent in multiple CDs
    		// loop over all the CDs that this agent is a member of
    		for (int i = 0; i < CDsJoined.size(); i++) {
    			URLDescriptor cd = CDsJoined.get(i);
    			subscribeViewsToCD(cd, view);
    		}
    	}
    	return msg_status;
    }
    
    NodeStyleSheet agentStyle = new NodeStyleSheet(null, NodeStyleSheet.DEFAULT_PROPERTIES);
    ArcStyleSheet arcStyle = new ArcStyleSheet(null, ArcStyleSheet.DEFAULT_PROPERTIES);
    
    NodeStyleSheet[] nodeStyles = { agentStyle };
    GeneralStyleSheet[] arcStyles = { arcStyle };
    
    /** Notify the view (if null, notify all views) of a CD and all it's subscribed members. */
    void subscribeViewsToCD(URLDescriptor cd, URLDescriptor view) {
    	// send a command to create CD node
    	Command addCD = comMaker.addNode(StringEscaper.escapeString(cd.toString()), getOtherAgentName(cd), 0, 0, 100, 100, nodeStyles, null);
    	addCD.setCommandArgument("modelID", StringEscaper.escapeString(cd.toString()));
    	if (view != null) {
    		sendCommand(addCD, view); // send command JUST to that view
    	}
    	else {
    		sendCommand(addCD);
    	}
    	
//  	send it a list of CD members
    	Vector members = getMembers(cd);
    	for (int j = 0; j < members.size(); j++) {
    		addAgent((URLDescriptor)members.get(j), cd, view); // just for that view
    	}
    }
    
    /** Commands the subscribed view agents to add a new agent that has subscribed to the CD. */
    void addAgent(URLDescriptor newMember, URLDescriptor cd, URLDescriptor view) {
    	askForName(newMember);
//  	make a new command that adds a node
    	Command addAgent = comMaker.addNode(StringEscaper.escapeString(newMember.toString()), getOtherAgentName(newMember), 0, 0, 100, 100, nodeStyles, null);
    	addAgent.setCommandArgument("modelID", StringEscaper.escapeString(newMember.toString()));
    	
    	// make a command that makes an arc between the CD and that node
    	// make an arc attached to a CD first:
    	Command makeArc = comMaker.addArc(ObjectID.newArc.toString(), "member", 0, 0, 100, 100, StringEscaper.escapeString(cd.toString()), arcStyles);
    	makeArc.setCommandArgument("modelID", StringEscaper.escapeString(newMember.toString() + "conArc"));
    	
    	// now attach that Arc to the new node
    	Command attach = comMaker.connectArcToObject(StringEscaper.escapeString(newMember.toString() + "conArc"), StringEscaper.escapeString(newMember.toString()), 2);
     	
    	if (view == null) {
     		sendCommand(addAgent);
    		sendCommand(makeArc);
    		sendCommand(attach);	
    	}
    	else {
    		sendCommand(addAgent, view);
    		sendCommand(makeArc, view);
    		sendCommand(attach, view);	
    	}
    }
    
    /** Commands the subscribed view agents to remove the representation
     *  of an agent that has withdrawn from the CD. */
    void removeAgent(URLDescriptor removedMember, URLDescriptor cd) {
    	// send command to remove node representation of the node agent
    	Command com = comMaker.removeNode(StringEscaper.escapeString(removedMember.toString()));
    	sendCommand(com);
    	
    	// remove arc connecting member to CD
    }
}
