/**
 * <p>Copyright: Copyright (c) 2005, Knowledge Science Group, University of Calgary.
 * Permission to use, copy, modify, distribute and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that copyright
 * notice and this permission notice appear in supporting documentation.  The
 * Knowledge Science Group makes no representations about the suitability of
 * this software for any purpose.  It is provided "as is" without express or
 * implied warranty.</p>
 */
package kartaCASA.OntologyEditor;

import casa.exceptions.IPSocketException;
import casa.interfaces.TransientAgentInterface;
import casa.util.TypeNode;
import casa.Status;

import kartaCASA.KartaViewAgent;

import karta.command.Command;
import karta.command.CommandMaker;
import karta.util.ObjectID;
import karta.Pickler.StringTypeWrapper;
import karta.stylesheets.*;

import java.util.Vector;
import java.util.Map;
import java.util.WeakHashMap;
import java.util.Queue;
import java.util.LinkedList;
import java.util.Collections;

/**
 * View agent for the ontology editor. This also doubles as the model agent. This is the agent
 * that is embedded inside the agent whose ontology is being edited - the actual window is a 
 * newly spawned Supervisor agent.
 *
 * Created on August 4, 2005
 * @author <a href="mailto:sedachv@cpsc.ucalgary.ca">Vladimir Sedach</a>
 * @version 0.13
 */
public class OntologyEditorView extends KartaViewAgent {
	
	boolean debugp = true;
	
    NodeStyleSheet nodeStyle = new NodeStyleSheet(null, NodeStyleSheet.DEFAULT_PROPERTIES);
    ArcStyleSheet arcStyle = new ArcStyleSheet(null, ArcStyleSheet.DEFAULT_PROPERTIES);
    
    NodeStyleSheet[] nodeStyles = { nodeStyle };
    GeneralStyleSheet[] arcStyles = { arcStyle };

	/** The agent whose ontology this view is editing. */
	protected TransientAgentInterface ownerAgent;
	
	protected CommandMaker comMaker;
	
	protected OntNode ontologyModelRoot;
	
	private void initCommonSlots() {
		commandFilter = new OntologyModelFilter();
		comMaker = new CommandMaker(pickler);
		ontologyModelRoot = null;
	}
	
	public OntologyEditorView(Qualifiers quals, TransientAgentInterface owner) throws IPSocketException {
		super(quals);

		initCommonSlots();
		if (debugp)
			System.out.println("OntologyEditorView finished initializing common slots.");
		setOwnerAgent(owner);
		if (debugp)
			System.out.println("OntologyEditorView finished setting owner agent.");
	}

	public TransientAgentInterface getOwnerAgent() {
		return ownerAgent;
	}

	public void setOwnerAgent(TransientAgentInterface ownerAgent) {
		this.ownerAgent = ownerAgent;
		readOwnerOntology();
	}
	
	protected void clearCurrentOntology() {
		if (ontologyModelRoot != null) {
			deleteNodeRepresentation(ontologyModelRoot);
			deleteNode(ontologyModelRoot);
			ontologyModelRoot = null;
		}
	}
	
	public void readOwnerOntology() {
		clearCurrentOntology();
		layOutTypeHierarchy();
		if (debugp)
			System.out.println("OntologyEditorView has finished laying out the type hierarchy.");
	}
	
	protected void notifyStateChangingCommand (Command com) {
		// the three types of commands that can affect the ontology are:
		// 1. remove node (which removes that node and all descendants with
		//    only that parent as node from the ontology).
		// 2. set name (renames a node).
		// 3. add node connected to nodes (add a node to the ontology).
		// in all three cases, we make those changes to the OntNode representation
		// then print it out to a string
		// then feed it into the ontology serialization interpreter
		// need to make some changes to the way the thing are drawn
		// only the 1st time the ontology is drawn should it be automatically laid out
		// afterwards, re-drawing of the graph should check what has and hasn't been drawn
		// and draw in only those parts that have changed, or maybe this even won't be necessary?

		if ((com.getType() == Command.REMOVE_NODE) || (com.getType() == Command.COLLAPSE_NODE)) {
			if (debugp)
				System.out.println("Received command to delete node:" + com);
			// find that node in the tree
			OntNode deleted = findNodeByID(ontologyModelRoot, new ObjectID((String)com.getCommandArgument("ID")));
			if (deleted != null) {
				if (debugp)
					System.out.println("Now will delete node " + deleted);
				deleteNodeRepresentation(deleted);
				if (com.getType() == Command.REMOVE_NODE)
					deleteNode(deleted);
				else if (com.getType() == Command.COLLAPSE_NODE)
					collapseNode(deleted);
			}
			else {
				System.out.println("Error: can't find node with ID " + com.getCommandArgument("ID") + " to delete.");
			}
		}
		else if (com.getType() == Command.ADD_NODE) {
			addNode(com);
		}
		else if (com.getType() == Command.SAVE_CHANGES) {
			String ontology = printOntology(ontologyModelRoot);
			
			if (debugp)
				System.out.println("Now saving ontology:\n" + ontology);
			
			Status stat = ownerAgent.putReplacementOntology(ontology);
			int status = stat.getStatus();
			
			if (status != 0) { // something's wrong
				System.out.println(stat);
			}
		}
	}
	
	private void addNode(Command com) {
		String name = ((StringTypeWrapper)com.getCommandArgument("name")).getString();
		int x = Integer.parseInt((String)com.getCommandArgument("left"));
		int y = Integer.parseInt((String)com.getCommandArgument("top"));
		
		Vector parents = (Vector)com.getCommandArgument("connectedNodes");
	
		OntNode newNode = new OntNode(name);
		// draw the node (assigns an ID while it's at it) 
		drawNode(newNode, x, y, 30); // !!!! MAGIC NUMBER ALERT !!!!
		// look at all the parents,
		for (int i = 0; i < parents.size(); i++) {
			OntNode parent = findNodeByID(ontologyModelRoot, new ObjectID((String)parents.get(i)));
			connectNodes(parent, newNode);
		}
	}
	
	private void connectNodes(OntNode parent, OntNode child) {
		// connect them in the ontology model
		parent.getChildren().add(child);
		child.getParents().add(parent);
		// add arcs between them
		drawArc(parent, child);
	}
	
	/**
	 * Utility function that removes a node from all of it's parents' childrens' lists
	 * and deletes the connecting arcs from the graph.
	 */
	private void deleteNodeFromParents(OntNode deleted) {
		// remove it from all of it's parents' childrens' lists
		for (int i = 0; i < deleted.getParents().size(); i++) {
			deleted.getParents().get(i).getChildren().remove(deleted);
		}

		// send command to delete it's parent arcs
		for (int i = 0; i < deleted.getParentArcIDs().size(); i++) {
			if (debugp)
				System.out.println("The number of arcs to parents of the node is " + deleted.getParentArcIDs().size());
			if (debugp)
				System.out.println("Now sending out command to delete arc attaching " + deleted + " to " + deleted.getParentArcIDs().get(i).parent);
			Command deleteArc = comMaker.removeArc(deleted.getParentArcIDs().get(i).ID);
			putCommandInQueue(deleteArc);
		}
	}
	
	/**
	 * Deletes the specified node from the graph, removing all arcs pointing from it to it's parents,
	 * and all descendant nodes and their connecting arcs which do not have other ancestors besides
	 * this node.
	 */
	private void deleteNode(OntNode deleted) {
		if (debugp)
			System.out.println("Now deleting node " + deleted);
//		 the command to delete the root node has already been sent back,
		// so we postpone it for children
		
		deleteNodeFromParents(deleted);
		
		// loop for all children of that node
		if (debugp)
			System.out.println("The node " + deleted + " has children " + deleted.getChildren());
		
		// Note that since the recursive call to deleteNode inside the following
		// loop will remove nodes from deleted's children vector, we have to make 
		// a copy upon which we can safely operate
		Vector<OntNode> childrenCopy = (Vector<OntNode>)deleted.getChildren().clone();
		
		for (int i = 0; i < childrenCopy.size(); i++) {
			OntNode child = childrenCopy.get(i);
			// if the child has no other parents, send command to delete that node and it's children recursively
			if (child.getParents().size() == 1) {
				if (debugp)
					System.out.println("Node " + child.getRootName() + " has not other parents but " + deleted.getRootName());
				deleteNodeRepresentation(child);
				deleteNode(child);
				child.getParents().remove(deleted);
			}
			else { // if the node has other parents, just delete that arcID that points to the deleted parent
				removeChild(deleted, child);
			}
			
		}
		
		// now all the necessary modifications to the ontologyModel have been done
		// can print it out and feed it back to the actual agent ontology mechanism
	}
	
	private void deleteNodeRepresentation (OntNode node) {
		Command deleteNode = comMaker.removeNode(node.getAssignedID().toString());
		putCommandInQueue(deleteNode);
	}
	
	private void removeChild(OntNode parent, OntNode child) {
		if (debugp)
			System.out.println("The node " + child + " has parents " + child.getParents());
		for (int j = 0; j < child.getParentArcIDs().size(); j++) {
			if (child.getParentArcIDs().get(j).parent == parent) {
				Command deleteArc = comMaker.removeArc(child.getParentArcIDs().get(j).ID);
				putCommandInQueue(deleteArc);
			}
		}
		parent.getChildren().remove(child);
		child.getParents().remove(parent);
	}
	
	/**
	 * Removes a node from the graph by "collapsing" it - children (immediate descendants) inherit
	 * from that node's parents (immediate ancestors) directly and the node itself goes away.
	 */
	private void collapseNode (OntNode node) {
		Vector<OntNode> parents = (Vector<OntNode>)node.getParents().clone();
		Vector<OntNode> children = (Vector<OntNode>)node.getChildren().clone();
		
		deleteNodeFromParents(node);
		deleteNodeRepresentation(node);
		
		for (int i = 0; i < children.size(); i++) {
			OntNode child = children.get(i);
			removeChild(node, child);
			for (int j = 0; j < parents.size(); j++) {
				OntNode parent = parents.get(j);
				if (parent.getChildren().indexOf(child) == -1) { // they're not already connected
					connectNodes(parent, child);
				}
			}
		}
	}
	
	private OntNode foundNode = null;
	
	final OntNode findNodeByID(final OntNode root, final ObjectID id) {
		NodeDoer finder = new NodeDoer() {
			public void doNode(OntNode node) {
				if (node.getAssignedID().equals(id)) {// if the node has not yet been linearized 
					foundNode = node;
				}
			}
		};

		depthFirstTraversal(finder, root);
		
		return foundNode;
	}
	
	/** Gets the root TypeNode of the ownerAgent's ontology. */
	protected TypeNode getRootNode() {
		if (ownerAgent != null) {
			if (debugp)
				System.out.println("The ontology root node looks like " + ownerAgent.getOntology().getRootNode());
			return ownerAgent.getOntology().getRootNode();
		}
		return null;
	}
	
	protected void layOutTypeHierarchy() {
		if (getRootNode() != null) {
			ontologyModelRoot = copyTypeHierarchy(getRootNode());
			
			if (debugp)
				System.out.println("Ontology looks like:\n" + printOntology(ontologyModelRoot));
			
			drawOntology(ontologyModelRoot);
		}
	}
	
	protected void depthFirstTraversal(NodeDoer doer, OntNode root) {
		if (root != null) {
			for (int i = 0; i < root.getChildren().size(); i++) {
				depthFirstTraversal(doer, root.getChildren().get(i));
			}
			doer.doNode(root);
		}
	}
	
	protected void breadthFirstTraversal(NodeDoer doer, OntNode root) {
		Queue<OntNode> nodeQueue = new LinkedList<OntNode>();
		nodeQueue.offer(root);
		while(nodeQueue.peek() != null) {
			OntNode node = nodeQueue.remove();
			doer.doNode(node);
			Vector<OntNode> children = node.getChildren();
			for (int i = 0; i < children.size(); i++) {
				nodeQueue.offer(children.get(i));
			}
		}
	}
	
	private Map<TypeNode, OntNode> dupe_table = new WeakHashMap<TypeNode, OntNode>();
	
	private OntNode getNode(TypeNode node) {
		if (dupe_table.get(node) != null) {
			return dupe_table.get(node);
		}
		OntNode newNode = new OntNode(node.name);
		dupe_table.put(node, newNode);
		return newNode;
	}		
	
	protected OntNode copyTypeHierarchy(TypeNode root) {
		OntNode treeRoot = getNode(root);
			
		for (int i = 0; i < root.getChildren().size(); i++) {
			OntNode child = copyTypeHierarchy((TypeNode)root.getChildren().get(i));
			
			if (child.getParents().contains(treeRoot)) {
				System.out.println("That's a bad TypeNode! Node " + child.getRootName() + " already has parent " + treeRoot.getRootName());
			}
			else {
				child.getParents().add(treeRoot);

			}
			if (treeRoot.getChildren().contains(child)) {
				System.out.println("That's a bad TypeNode! Node " + treeRoot.getRootName() + " already has parent " + child.getRootName());
			}
			else {
				treeRoot.getChildren().add(child);

			}
			
		}
		return treeRoot;
	}
	
	/** 
	 * Assigns levels to all nodes in the graph, placing each node below it's lowest parents
	 * @return Returns the depth of the graph
	 */
	protected int assignLevels(OntNode root) {
		int nextLevel = root.getLevel() + 1;
		int maxLevel = nextLevel;
		for (int i = 0; i < root.getChildren().size(); i++) {
			OntNode child = root.getChildren().get(i);
			int oldChildLevel = child.getLevel();
			if (oldChildLevel <= nextLevel) {
				child.setLevel(nextLevel);
				int childLevel = assignLevels(child);
				if (childLevel > maxLevel)
					maxLevel = childLevel;
			}
		}
		return maxLevel;
	}
	
	/** Utility function for barycenterGraphSort that fills in a matrix representation
	 * of the given parent-child subgraph.
	 */
	private void fillConnectionsMatrix(boolean[][] connections, Vector<OntNode> parents, Vector<OntNode> children) {
		int numChildren = children.size();
		int numParents = parents.size();
		
		for (int j = 0; j < numChildren; j++) {
			for (int k = 0; k < numParents; k++) {
				connections[k][j] = false;
			}
		}
		for (int j = 0; j < numParents; j++) {
			for (int k = 0; k < parents.get(j).getChildren().size(); k++) {
				if (children.indexOf(parents.get(j).getChildren().get(k)) != -1)  // if the current child node is part of the children set
					connections[j][children.indexOf(parents.get(j).getChildren().get(k))] = true;
			}
		}
	}
	
	/** Sorts a grid representation of a graph according to the 
	 * barycenter heuristic to minimize number of edge crossings. 
	 */
	protected void barycenterGraphSort(Vector<Vector<OntNode>> grid) {
		for (int i = 1; i < grid.size(); i++) {
			// construct a matrix representing connections
			Vector<OntNode> parents = grid.get(i - 1);
			Vector<OntNode> children = grid.get(i);
			int numParents = parents.size();
			int numChildren = children.size();
			boolean[][] connections = new boolean[numParents][numChildren];
			
			fillConnectionsMatrix(connections, parents, children);
			
			// computer row barycenters
			for (int j = 0; j < numParents; j++) {
				int posTally = 0;
				int numConnections = 0;
				for (int k = 0; k < numChildren; k++) {
					if (connections[j][k]) {
						posTally += k;
						numConnections++;
					}
				}
				parents.get(j).setBarycenter((numConnections == 0) ? 0 : posTally / numConnections);
			}
			
			// sort parent nodes by row barycenters
			Collections.sort(parents);
			
			
			// update matrix
			fillConnectionsMatrix(connections, parents, children);
			
			// compute column barycenters
			for (int j = 0; j < numChildren; j++) {
				int posTally = 0;
				int numConnections = 0;
				for (int k = 0; k < numParents; k++) {
					if (connections[k][j]) {
						posTally += k;
						numConnections++;
					}
				}
				children.get(j).setBarycenter(posTally / numConnections);
			}
			// sort child nodes by column barycenters
			Collections.sort(children);
		}
	}
	
	private Vector<Vector<OntNode>> copyGrid(Vector<Vector<OntNode>> grid) {
		Vector<Vector<OntNode>> copy = new Vector<Vector<OntNode>>();
		for (int i = 0; i < grid.size(); i++) {
			copy.add((Vector<OntNode>)grid.get(i).clone());
		}
		return copy;
	}
	
	/**
	 * Linearize the graph as a grid of nodes according to their level
	 */
	private Vector<Vector<OntNode>> gridLinearize(OntNode root, int graphDepth) {
		final Vector<Vector<OntNode>> grid = new Vector<Vector<OntNode>>();
		
		// initialize grid arrays
		for (int i = 0; i < graphDepth; i++) {
			grid.add(new Vector<OntNode>());
		}

		// A list that keeps track of which nodes have been added to the grid
		final Map<OntNode, Boolean> visitedNodes = new WeakHashMap<OntNode, Boolean>();
		
		NodeDoer linearizer = new NodeDoer() {
			public void doNode(OntNode node) {
				if (visitedNodes.get(node) == null) // if the node has not yet been linearized 
					grid.get(node.getLevel()).add(node);
				visitedNodes.put(node, true);
			}
		};
		
		// Do the actual linearization
		breadthFirstTraversal(linearizer, root);
		
		return grid;
	}
	
	protected void drawNode(OntNode node, int xPos, int yPos, int nodeSize) {
		ObjectID nodeID = pickler.getNewNodeID();
		node.setAssignedID(nodeID);

		Command makeNode = comMaker.addNode(nodeID, node.getRootName(), xPos, yPos, nodeSize, nodeSize, nodeStyle, null);
		putCommandInQueue(makeNode);
	}
	
	protected void drawArc(OntNode parent, OntNode child) {
		ObjectID parentID = parent.getAssignedID();
		ObjectID childID = child.getAssignedID();
		ObjectID arcID = pickler.getNewArcID();
		
		Command addArc = comMaker.addArc(arcID.toString(), "", 0, 0, 0, 0, parentID.toString(), arcStyles);
		putCommandInQueue(addArc);
		Command connectNode = comMaker.connectArcToObject(arcID.toString(), childID.toString(), 2);
		putCommandInQueue(connectNode);
		
		child.getParentArcIDs().add(new arcID(parent, arcID));
	}
	
	protected void drawOntology(OntNode root) {
		// constant representing the diameter of each node
		final int nodeSize = 30;
		
		// Assign levels to graph nodes and compute graph depth
		final int graphDepth = assignLevels(root);
		
		if (debugp)
			System.out.println("Graph depth is " + graphDepth);
		
		final Vector<Vector<OntNode>> grid = gridLinearize(root, graphDepth);

		// determine maximum width of graph (which level has the most nodes)
		int maxGraphWidth = 0;
		for (int i = 0; i < grid.size(); i++) {
			if (grid.get(i).size() > maxGraphWidth)
				maxGraphWidth = grid.get(i).size();
		}

		// sort horizontal position of nodes to minimize edge crossings
		barycenterGraphSort(grid);
		
		// determine x coordinates: center, minimum left, node offset
		// maxPixelWidth is the width of the wides row on nodes in pixels
		int maxPixelWidth = 0;
		
		// levelXNodeDistance is an array that holds the distance between two nodes, indexed by the level of the graph
		final int[] levelXNodeDistance = new int[graphDepth];
		for (int i = 0; i < graphDepth; i++) {
			int levelDistance = (int)(nodeSize * 1.7) + (int)(0.06 * grid.get(i).size() * nodeSize);
			
			levelXNodeDistance[i] = levelDistance;
			
			int currentPixelWidth = (levelDistance * grid.get(i).size());
			
			if (maxPixelWidth < currentPixelWidth)
				maxPixelWidth = currentPixelWidth;
		}
		
		// now go over backwards to "fatten up" the graph
		for (int i = graphDepth - 1; i > 2; i--) {
			int currentPixelWidth = (levelXNodeDistance[i] * grid.get(i).size());
			
			int previousLevelWidth = levelXNodeDistance[i-1] * grid.get(i-1).size();
			int fractionCurrentWidth = (int)(0.9 * currentPixelWidth);
			
			if (previousLevelWidth < fractionCurrentWidth)
				levelXNodeDistance[i-1] = fractionCurrentWidth / grid.get(i-1).size();
		}
		
		// pixelCenter is where the center line of the tree will be
		int pixelCenter = maxPixelWidth / 2;
		
		// levelXOffset is an array, indexed by graph level, of the pixel coordinate from which a node row should be begun to be drawn
		final int[] levelXOffset = new int[graphDepth];
		for (int i = 0; i < graphDepth; i++) {
			int levelWidth = grid.get(i).size();
			int xoffset = pixelCenter - (((levelWidth + 1) * levelXNodeDistance[i]) / 2);
			levelXOffset[i] = xoffset;
		}
		
		// levelYPosition is an array, indexed by graph level, of the y-coordinate of each level of the graph
		final int[] levelYPosition = new int[graphDepth];
		for (int i = 0; i < graphDepth; i++) {
			int levelYOffset = (int)(grid.get(i).size() * nodeSize * .17);
			levelYPosition[i] = ((i == 0) ? 0 : levelYPosition[i-1]) + levelYOffset + (nodeSize * 2);
		}
		
		// draw nodes from grid (which has no duplicates since it is derived from the linearization)
		for (int i = 0; i < grid.size(); i++) {
			for (int j = 0; j < grid.get(i).size(); j++) {
				OntNode node = grid.get(i).get(j);
				int xPos = (levelXOffset[i] += levelXNodeDistance[i]);
				int yPos = levelYPosition[i]; 
 				drawNode(node, xPos, yPos, nodeSize);
			}
		}
		
		// now to connect the nodes with arcs
		for (int i = 0; i < grid.size(); i++) {
			for (int j = 0; j < grid.get(i).size(); j++) {
				OntNode currentNode = grid.get(i).get(j);
				Vector<OntNode> children = currentNode.getChildren();
				
				for (int k = 0; k < children.size(); k++) {
					drawArc(currentNode, children.get(k));
				}
			}
		}
	}
		
	/**
	 * Nested class used as a closure to operate on TypeNodes by different TypeNode tree
	 * traversal functions.
	 */
	class NodeDoer {
		public void doNode(OntNode node) {
			
		}
	}
	
	/** A parent-arc tuple that associates a given arcID with the parent node to which it is attached. */
	class arcID {
		public OntNode parent;
		public ObjectID ID;
		
		public arcID(OntNode p, ObjectID a) {
			parent = p;
			ID = a;
		}
	}
	
	/**
	 * A convenient and compact representation for the ontology graph.
	 */
	class OntNode implements java.lang.Comparable {
		String rootName;
		Vector<OntNode> parents;
		Vector<OntNode> children;
		int level;
		float barycenter;
		
		ObjectID assignedID;
		Vector<arcID> parentArcIDs;
		
		public OntNode(String name) {
			children = new Vector<OntNode>();
			parents = new Vector<OntNode>();
			setRootName(name);
			level = 0;
			
			parentArcIDs = new Vector<arcID>();
		}
		
		public int compareTo(Object obj) {
			if (obj instanceof OntNode) {
				if (barycenter < ((OntNode)obj).getBarycenter())
					return -1;
				else if (barycenter > ((OntNode)obj).getBarycenter())
					return 1;
			}
			return 0;
		}
		
		public OntNode clone() {
			OntNode root = new OntNode(getRootName());
			Vector<OntNode> childrenCopy = new Vector<OntNode>();
			for (int i = 0; i < getChildren().size(); i++) {
				childrenCopy.add(getChildren().get(i).clone());
			}
			root.setChildren(childrenCopy);
			Vector<OntNode> parentCopy = new Vector<OntNode>();
			for (int i = 0; i < getParents().size(); i++) {
				parentCopy.add(getParents().get(i).clone());
			}
			root.setParents(parentCopy);
			root.setParentArcIDs((Vector<arcID>)parentArcIDs.clone()); 
			return root;
		}

		public Vector<OntNode> getChildren() {
			return children;
		}

		public String getRootName() {
			return rootName;
		}

		public void setRootName(String rootName) {
			this.rootName = rootName;
		}

		public void setChildren(Vector<OntNode> descendants) {
			this.children = descendants;
		}

		public int getLevel() {
			return level;
		}

		public void setLevel(int level) {
			this.level = level;
		}

		public ObjectID getAssignedID() {
			return assignedID;
		}

		public void setAssignedID(ObjectID assignedID) {
			this.assignedID = assignedID;
		}

		public float getBarycenter() {
			return barycenter;
		}

		public void setBarycenter(float barycenter) {
			this.barycenter = barycenter;
		}

		public Vector<OntNode> getParents() {
			return parents;
		}

		public void setParents(Vector<OntNode> parents) {
			this.parents = parents;
		}

		public Vector<arcID> getParentArcIDs() {
			return parentArcIDs;
		}

		public void setParentArcIDs(Vector<arcID> parentArcIDs) {
			this.parentArcIDs = parentArcIDs;
		}
		
		public String toString() {
			String lineString = karta.util.StringEscaper.escapeString(getRootName()) + " <";
			for (int i = 0; i < parents.size(); i++) {
				lineString = lineString + " " + karta.util.StringEscaper.escapeString(parents.get(i).getRootName());
			}
			lineString = lineString + ";";
			return lineString;
		}
	}
	
	/// Functions for saving the OntNode ontology model
	private Map<OntNode, Object> printedNodes = new WeakHashMap<OntNode, Object>();
	
	private boolean allParentsPrintedP (OntNode node) {
		for (int i = 0; i < node.getParents().size(); i++) {
			if (printedNodes.get(node.getParents().get(i)) == null)
				return false;
		}
		return true;
	}
	
	public String printOntology (OntNode root) {
		final StringBuffer ontology = new StringBuffer();
		
		NodeDoer printer = new NodeDoer() {
			public void doNode(OntNode node) {
				if ((printedNodes.get(node) == null) && allParentsPrintedP(node)) {
					ontology.append(node.toString());
					ontology.append("\n");
					printedNodes.put(node, node);
				}
			}
		};
		
		breadthFirstTraversal(printer, root);
		
		ontology.append(";");
		
		printedNodes.clear();
		return ontology.toString();
	}
}

