/**
 * <p>Title: CASA Agent Infrastructure</p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2007, Knowledge Science Group, University of Calgary. 
 * Permission to use, copy, modify, distribute and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that copyright
 * notice and this permission notice appear in supporting documentation.  
 * The  Knowledge Science Group makes no representations about the suitability
 * of  this software for any purpose.  It is provided "as is" without express
 * or implied warranty.</p>
 * <p>Company: Knowledge Science Group, University of Calgary</p>
 * @author <a hef="http://pages.cpsc.ucalgary.ca/~kremer/">Rob Kremer</a>
 * @version 0.17
 */
package casa.extensions;

import casa.TransientAgent;
import casa.ui.AbstractInternalFrame;
import casa.ui.AgentUI;
import casa.util.CASAUtil;
import casa.util.JarLoader;
import casa.util.Runnable1;

import java.awt.Component;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.jar.Attributes;
import java.util.jar.JarFile;
import java.util.jar.Manifest;

import javax.swing.JComponent;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.ScrollPaneConstants;

/**
 * @author <a hef="http://pages.cpsc.ucalgary.ca/~kremer/">Rob Kremer</a>
 *
 */
public class TabJar {
	TransientAgent agent;
	AbstractInternalFrame frame;
	String name;
	boolean visible;
	File jarFile;
	String className;
	Class<?>[] constructorTypes;
	Object[] constructorObjects;
	public JComponent component = null;
	
	/**
	 * Creates a "lazy" tab for an agent.
	 * @param agent The agent who "owns" the tab.
	 * @param frame The frame in the interface to which the tab should be attached.
	 * @param name The label on the tab.
	 * @param visible if true the tab will be instantiated, otherwise the tab will be uninstantiated but on the tab menu "Windows|Tabs|[name]". 
	 * @param jarPath The path to the jar in which the extension resides.
	 * @param className The full path name if the class that implements the extension.
	 * @param constructorTypes The types of the constructor parameters.
	 * @param constructorObjects The actual objects to use in calling the constructor.
	 * @throws IOException 
	 */
	public TabJar(TransientAgent agent, 
								AbstractInternalFrame frame, 
								String name, 
								boolean visible,
								File jarPath, 
								String className, 
								Class<?>[] constructorTypes, 
								Object[] constructorObjects) 
										throws IOException {
		assert agent!=null;
		assert frame!=null;
		assert name!=null;
		assert jarPath!=null;
		assert className!=null;
		this.agent = agent;
		this.frame = frame;
		this.name = name;
		this.visible = visible;
		this.className = className;
		this.constructorTypes = constructorTypes;
		this.constructorObjects = constructorObjects;
		jarFile = jarPath;
		if (!jarFile.exists()) {
			throw new IOException("Jar file "+jarFile+" does not exist.");
		}
		initializeTab();
	}
	
	public static void main(String[] args) {
		Runnable1<TransientAgent,Void> code = new Runnable1<TransientAgent,Void>() {
			@Override
			public Void run(TransientAgent agent) {
				try {
				AgentUI ui = agent.getPrimaryUI();
				AbstractInternalFrame aif = (AbstractInternalFrame)ui;
				File file = new File("CasaUMLetFrame.jar");
				new TabJar(agent, aif, "UMLet", false, file, "casaUmlet.CasaUMLetFrame", (Class<?>[])null, (Object[])null);
				} catch (Throwable e) {
					e.printStackTrace();
				}
				return null;
			}
		};
		CASAUtil.startAnAgent(TransientAgent.class, "Fred4", 6234, code);
	}
	
	/**
	 * Looks for jars that qualify as tab components in ~/.casa/extensions and instantiates "lazy" tabs
	 * from the jars.  A qualifying jar contains a manifest file with sections named "tabs0", "tabs1"... with the
	 * following attributes:
	 * <ul>
	 * <li> Main: (required) The fully-qualified class name in the jar file that defines a subtype
	 * of {@link java.awt.Component} with a constructor with two parameters:
	 * ({@link TransientAgent}, {@link AbstractInternalFrame}).  This will be invoked 
	 * as the main component in the tab.
	 * <li> TabName: (optional) The name to appear on the tab itself. The default is class name
	 * given in Main (with no package names).
	 * <li> AgentType: (optional) The fully qualified class name of that restricts the 
	 * type of the agent this tab will apply to.  The default is "casa.TransientAgent".
	 * <li> InterfaceType: (optional) The fully qualified class name that restricts the 
	 * type of the interface this tab will apply to.  The default is "casa.ui.AbstractInternalFrame".
	 * <li> Visible: (optional) if "true" the jar WILL be loaded and tab will be visible. The default is "false". 
	 * </ul>
	 * Note that by default, the jar will NOT be loaded and tabs will NOT be displayed when the 
	 * interface is inialized.  To display the tab interface use the menu bar "Windows | Tabs | [tabName]".
	 * To actually load the jar and run the tab code, choose the tab to display it -- the jar will
	 * only be loaded and run at that point.
	 * @param agent The agent who "owns" the interface.
	 * @param frame The frame that contains the tab.
	 * @throws Exception
	 */
	public static void lookupJars(TransientAgent agent, AbstractInternalFrame frame) throws Exception {
		try {
		assert agent!=null;
		assert frame!=null;
		File dir = new File(System.getProperty("user.home") + File.separator + ".casa"+ File.separator + "extensions");
		if (!dir.isDirectory())
			return;
		File[] files = dir.listFiles(new FileFilter() {
			@Override
			public boolean accept(File arg0) {
				// TODO Auto-generated method stub
				return arg0.getName().endsWith(".jar");
			}
		});
		for (File file: files) {
			JarFile jarFile = new JarFile(file);
			Manifest manifest = jarFile.getManifest();
			Attributes attr;
			String tabNum;
			for (int i=0; (attr=manifest.getAttributes(tabNum=("tabs"+i)))!=null; i++) {
				String agentType = attr.getValue("AgentType");
				if (agentType!=null) {
					try {
						Class<?> cls = Class.forName(agentType);
						if (!cls.isAssignableFrom(agent.getClass()))
							continue;
					} catch (Exception e) {
						agent.println("warning", "invalid class name '"+agentType+"' in the 'agentType' attribute in jar '"+file+"', manifest section '"+tabNum+"'.  Ignoring it.", e);
					}
				}
				String interfaceType = attr.getValue("InterfaceType");
				if (interfaceType!=null) {
					try {
						Class<?> cls = Class.forName(interfaceType);
						if (!cls.isAssignableFrom(frame.getClass()))
							continue;
					} catch (Exception e) {
						agent.println("warning", "invalid class name '"+interfaceType+"' in the 'interfaceType' attribute in jar '"+file+"', manifest section '"+tabNum+"'.  Ignoring it.", e);
					}
				}
				String obj = attr.getValue("Visible");
				boolean visible = false;
				if (obj!=null && obj.equalsIgnoreCase("true")) {
					visible = true;
				}
				String mainClassString = attr.getValue("Main");
				if (mainClassString==null)
					throw new Exception("Expected a 'Main' attribute in the jar '"+file+"', manifest section '"+tabNum+"'");
				String nameString = attr.getValue("TabName");
				if (nameString==null) {
					nameString = mainClassString.substring(mainClassString.lastIndexOf('.')<0?0:mainClassString.lastIndexOf('.')+1);
				}
				try {
					new TabJar(agent, frame, nameString, visible, file, mainClassString, (Class<?>[])null, (Object[])null);
				} catch (Throwable e) {
					throw new Exception("Failed to create new TabJar object for entry '"+mainClassString+"' in jar '"+file+"'in jar '"+file+"', manifest section '"+tabNum+"'", e);
				}
			}
		}
		}
		catch (Throwable e) {
			agent.println("error", "TabJar.lookupJars(): unexpected exception", e);
		}
	}
	
	/**
	 * Loads the jar (specified in the constructor) 
	 * @throws IOException
	 * @throws ClassCastException
	 * @throws ClassNotFoundException
	 * @throws SecurityException
	 * @throws NoSuchMethodException
	 * @throws IllegalArgumentException
	 * @throws InstantiationException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	public void loadJar() throws IOException, ClassCastException, ClassNotFoundException, SecurityException, NoSuchMethodException, IllegalArgumentException, InstantiationException, IllegalAccessException, InvocationTargetException {
	//	JarLoader.
		JarLoader.addFile(jarFile);
		if (constructorTypes==null) {
			constructorTypes = new Class[]{TransientAgent.class, AbstractInternalFrame.class};
			constructorObjects = new Object[]{agent, frame};
		}
		Class<?> cls = Class.forName(className);
		Constructor<?> constructor = cls.getConstructor(constructorTypes);
		Object o = constructor.newInstance(constructorObjects);
		if (o instanceof Component) {
			component = (JComponent)o;
		}
		else {
			throw new ClassCastException("Tab plugins must specify a subclass of JComponent, but "+className+" is of type "+o.getClass().getCanonicalName());
		}
		frame.replaceTabComponent(name, component, null);
		frame.setSelectedTab(name);
	}

	private boolean isInitialized = false;
	
	/**
	 * Adds a new tab to the frame (as specified in the constructor) but does not load the 
	 * jar; instead a listener is installed to load the jar (via {@link #loadJar()}) only when 
	 * the component of the tab becomes visible.  
	 */
	public void initializeTab() {
		component = new JTextArea("Please wait while the "+name+" plugin initializes...");
		JScrollPane scroll = new JScrollPane(component, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		ComponentListener listener = new ComponentListener() {
			@Override
			public void componentShown(ComponentEvent e) {
				if (!isInitialized) {
					try {
						loadJar();
						isInitialized = true;
					} catch (Throwable e1) {
						((JTextArea)component).setText(CASAUtil.log("error", "Could not load plugin", e1, true));
					}
				}
			}
			@Override
			public void componentResized(ComponentEvent e) {
			}
			@Override
			public void componentMoved(ComponentEvent e) {
			}
			@Override
			public void componentHidden(ComponentEvent e) {
			}
		};
		scroll.addComponentListener(listener);
		component.addComponentListener(listener);
		frame.addTab(name, scroll, visible);
	}
}
