package casa.util;

import java.io.ByteArrayOutputStream;
import java.io.PrintWriter;
import java.util.TreeSet;
import java.util.Vector;

import org.ksg.casa.CASA;

public abstract class TraceStatic implements TraceInterface {

	/** For platform-independence it's recommended that line.separator be used */
	protected static String LINE_SEPARATOR = System.getProperty("line.separator","\n");

	/** 
	 * WARNING: the timestamp format should NOT contain any colons (":"), as that would screw up any code that parses the header
	 * @see #getHeader(String);
	 */
	protected static String dateFormat = LONG_TIMESTAMP_FORMAT;
	
	/**
	 * An Trace object used by the static methods to track messages not associated with any particular agent.
	 */
	protected static Trace processWideTrace = null; 
	
	/**
	 * Logs the message to the TraceInterface object in the first argument.  If the first parameter
	 * is null, the log message will go to the process-wide Trace object.
	 * @param agent
	 * @param tag
	 * @param message
	 * @param ex
	 * @param options
	 * @return
	 */
	public static String log(TraceInterface agent, String tag, String message, Throwable ex, int options) {
		if (agent!=null) {
			return agent.println(tag, message, ex, options);
		}
		
		agent = CASAUtil.getAbstractProcessInScopeSilent();
		if (agent!=null) {
			return agent.println(tag, message, ex, options);
		}
		
		if (processWideTrace!=null)
			return processWideTrace.println(tag, message, ex, options);

		// TODO fix this to print something more appropriate.
		Pair<String,String> msg = format(tag, message, ex, options, "CASA: ", "");
		System.err.println(msg.getFirst());
		if (msg.getSecond()!=null)
			System.err.println(msg.getSecond());
		return message;
	}

	/**
	 * Logs the message if tag is in effect to the in-scope agent's log files, windows, etc.
	 * If no agent can be found in scope, uses the process-wide Trace object to perform the logging.
	 * @param tag A trace tag: if an agent can't be found in scope this is ignored
	 * @param message The message to log
	 * @param ex If ex is not null, the stack dump is appended to the message.
	 * @return The String (minus the generated log header) printed to the log
	 * @see Trace#log(String, String)
	 * @see CASAUtil#getAbstractProcessInScope()
	 */
	public static String log(String tag, String message, Throwable ex) {
		return log(null, tag, message, ex, 0);
	}
	
	/**
	 * Logs the message if tag is in effect to the in-scope agent's log files, windows, etc.
	 * If no agent can be found in scope, uses the process-wide Trace object to perform the logging.
	 * @param tag A trace tag: if an agent can't be found in scope this is ignored
	 * @param message The message to log
	 * @param ex If ex is not null, the stack dump is appended to the message.
	 * @return The String (minus the generated log header) printed to the log
	 * @see Trace#log(String, String)
	 * @see CASAUtil#getAbstractProcessInScope()
	 */
	public static String log(String tag, String message, Throwable ex, int options) {
		return log(null, tag, message, ex, options);
	}
	
	/**
	 * Logs the message if tag is in effect to the in-scope agent's log files, windows, etc.
	 * If no agent can be found in scope, does the best-effort to log to System output.
	 * @param tag A trace tag: if an agent can't be found in scope this is ignored
	 * @param message The message to log
	 * @return The String (minus the generated log header) printed to the log
	 * @see log
	 * @see CASAUtil#getAbstractProcessInScope()
	 */
	public static String log(String tag, String message) {
		return log(null, tag, message, null, 0);
	}

	/**
	 * Logs the message if tag is in effect to the in-scope agent's log files, windows, etc.
	 * If no agent can be found in scope, does the best-effort to log to System output.
	 * @param tag A trace tag: if an agent can't be found in scope this is ignored
	 * @param message The message to log
	 * @return The String (minus the generated log header) printed to the log
	 * @see log
	 * @see CASAUtil#getAbstractProcessInScope()
	 */
	public static String log(String tag, String message, int options) {
		return log(null, tag, message, null, options);
	}
	
	/**
	 * Converts the parameter Throwable to a string, including the stack trace and the "caused by" trace.
	 * @param ex
	 * @return The string version of the Throwable with the stack trace.
	 */
  public static String getStackTraceString(Throwable ex) {
    ByteArrayOutputStream out = new ByteArrayOutputStream();
    PrintWriter pw = new PrintWriter(out);
    for (Throwable th = ex; th!=null; th = th.getCause()) {
    	if (th!=ex) pw.println("Caused by:");
      th.printStackTrace(pw);
    }
    pw.println();
    pw.flush();
    return out.toString();
  }

  protected static Pair<String,String> format(String tag, String string, Throwable e, int flags, String header, String prefix) {
  	String stackTrace = null;
  	if (e!=null && (OPT_SUPRESS_STACK_TRACE&flags)==0)
  		stackTrace = getStackTraceString(e);
  	else if (e==null && (OPT_FORCE_STACK_TRACE&flags)!=0)
  		stackTrace = getTrimmedStackTraceString();

  	//set up flagging (>>>>>  <<<<< surrounds)
  	boolean flagging=false;
  	String inflagString=null, outflagString=null;
  	if (tag!=null) {
  		if (tag.length()>=5 && tag.substring(0,5).equals("error")) {
//  			errors++;
  			flagging = true;
  			inflagString  = ">>>>>>>>>>>>>>>>>>>>>>>>>";
  			outflagString = "<<<<<<<<<<<<<<<<<<<<<<<<<";
  			if (string.length()==0 || string.charAt(string.length()-1)!='\n') outflagString = "\n"+outflagString;
  		}
  		else if (tag.length()>=7 && tag.substring(0,7).equals("warning")) {
//  			warnings++;
  			flagging = true;
  			inflagString  = ">>>>>>>>>>>>>";
  			outflagString = "<<<<<<<<<<<<<";
  			if (string.length()==0 || string.charAt(string.length()-1)!='\n') outflagString = "\n"+outflagString;
  		}
  		else if ((flags&OPT_INCLUDE_CODE_LINE_NUMBER)!=0) {
  			flagging = true;
  			inflagString  = ">>>>>>>";
  			outflagString = "<<<<<<<";
  			if (string.length()==0 || string.charAt(string.length()-1)!='\n') outflagString = "\n"+outflagString;
  		}
  	}
  	if (flagging) {
  		String methodDescription = "unknown method";
  		StackTraceElement[] trace = getTrimmedStackTrace();
  		if (trace!=null && trace.length>0) { // we can't be certain how deep the "real" calling method is, but let's assume it's the first one that doesn't start with "print" :)
  			methodDescription = trace[0].toString();
  		}
  		inflagString += (" "+methodDescription+LINE_SEPARATOR);
  	}

  	StringBuilder buff = new StringBuilder();

  	if (flagging) 
  		buff.append(inflagString).append(LINE_SEPARATOR);
  	buff.append(header);
  	if (prefix != null) 
  		buff.append(prefix);
  	buff.append(string.replaceAll("\n",LINE_SEPARATOR));
  	if (stackTrace!=null)
  		buff.append(LINE_SEPARATOR).append(stackTrace);
  	if (flagging) 
  		buff.append(LINE_SEPARATOR).append(outflagString);
  	buff.append(LINE_SEPARATOR);
  	String msg = buff.toString();
  	
  	return new Pair<String,String>(msg, stackTrace);
  }


	/**
	 * The set of knowTags, initialized by {@link #getKnownTagsFromPersistentStore()}.
	 */
	private static TreeSet<String> knownTags = getKnownTagsFromPersistentStore();
	
	/**
	 * @return the set of all known tags, as specified by {@link CASA#getPreference(String, boolean, int) CASA.getPreference("knownTraceTags", "", 0)}
	 * @see #saveKnownTagsToPersistentStore()
	 * @see #updateTags(String)
	 */
	private static synchronized TreeSet<String> getKnownTagsFromPersistentStore() {
		String tags = CASA.getPreference("knownTraceTags", "", 0);
		String[] tagArray = tags.split(",");
		TreeSet<String> set = new TreeSet<String>();
		for (String t: tagArray) set.add(t);
		return set;
	}
	
	/**
	 * Saves the known tags back to persistent store by calling {@link CASA#putPreference(String, java.util.Map, int) CASA.putPreference("knownTraceTags", tags, CASA.USER|CASA.SYSTEM)}
	 * @see #getKnownTagsFromPersistentStore()
	 * @see #updateTags(String)
	 */
	private static synchronized void saveKnownTagsToPersistentStore() {
		String tags = getKnownTagsString();
		CASA.putPreference("knownTraceTags", tags, CASA.USER|CASA.SYSTEM);
	}
	
	/**
	 * Records the tag to persistent tags if it's not already there.
	 * @param tag The tag NAME (without trailing digit).
	 * @see #getKnownTagsFromPersistentStore()
	 * @see #saveKnownTagsToPersistentStore()
	 */
	protected static synchronized void updateTags(String tag) {
		if (tag!=null && tag.length()>0 && !knownTags.contains(tag)) {
			knownTags.add(tag);
			saveKnownTagsToPersistentStore();
		}
	}

	/**
	 * @return The known tags as a comma-delimited string; this method only uses the locally-cached known tags.
	 * @see #saveKnownTagsToPersistentStore()
	 */
	public static synchronized String getKnownTagsString() {
		if (knownTags.size()==0) return ""; 
		StringBuilder s = new StringBuilder();
		for (String tag: knownTags) s.append(',').append(tag);
		return s.substring(1);
	}
	
	/**
	 * Sets the set of known tags to the empty set, and stores it persistently (and globally to CASA).
	 */
	public static synchronized void clearKnownTags() {
		knownTags.clear();
		saveKnownTagsToPersistentStore();
	}
	
	/**
	 * @param t
	 * @return True iff the trace element t is part of the printing methods at top of stack; that is, the method name begins with "print", "display", or "log". 
	 */
	protected static boolean isPrintMethodElement(StackTraceElement t) {
		String methodName = t.getMethodName().toLowerCase(); 
		return methodName.startsWith("print") || methodName.startsWith("display")|| methodName.startsWith("log");
	}
	
	protected static StackTraceElement[] getTrimmedStackTrace() {
		return getTrimmedStackTrace(3);
	}

	/**
	 * @param skip The number of top-level element to skip
	 * @return The current thread's stack with all the print-type methods trimmed off the top.
	 * @see Trace#isPrintMethodElement(StackTraceElement)
	 */
	protected static StackTraceElement[] getTrimmedStackTrace(int skip) {
		Vector<StackTraceElement> ret = new Vector<StackTraceElement>();
		StackTraceElement[] trace = Thread.currentThread().getStackTrace();
		boolean trimming = true;
		if (trace!=null) { // we can't be certain how deep the "real" calling method is, but let's assume it's the first one that doesn't start with "print" :)
			for (StackTraceElement t: trace) {
				if (skip-- > 0) // skip the first elements, according to "skip"
					continue;
				if (trimming && isPrintMethodElement(t))
					continue;
				trimming = false;
				ret.add(t);
			}
		}
		return ret.toArray(new StackTraceElement[ret.size()]);
	}
	
	/**
	 * Returns the stack with the best guess at trimming the the Trace methods off the top.
	 * @return The trimmed stack formated as a String with tabs before each line.
	 * @see Trace#getTrimmedStackTrace()
	 * @see Trace#isPrintMethodElement(StackTraceElement)
	 */
	public static String getTrimmedStackTraceString() {
		StringBuilder buf = new StringBuilder();
		for (StackTraceElement t: getTrimmedStackTrace(3)) {
			buf.append("\n\t").append(t.toString());
		}
		String ret = buf.toString();
		if (ret.length()>0)
			ret = ret.substring(1);
		return ret;
	}
}
